/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "MeshVInfo.hpp"

namespace parncl{

MeshVInfo::MeshVInfo(ParCALVarSHnd *pv_shnd)
{
	size_t i=0;

	mesh_ = pv_shnd->get_mesh();
	mb_core_ = mesh_->get_mb_core();
	data_type_ = pv_shnd->get_data_type();

	moab::EntityType ent_type = ParCALLocToMB(pv_shnd->get_pcvar()->get_loc());
	moab::DataType dt = PNclNCMapToMB(data_type_);
	var_chunk_sz_ = PNclMBGetChunkSz(mesh_, pv_shnd->get_pcvar()->name().c_str(), dt, ent_type);

	pcvar_loc_ = pv_shnd->get_pcvar()->get_loc();
	pv_shnd->get_missing_val(pcvar_missing_val_);

	const std::vector<pargal::pcdim> &pcvar_dims = pv_shnd->get_pcvar()->get_dims();
	pcvar_dims_.resize(pcvar_dims.size());
	std::copy(pcvar_dims.begin(), pcvar_dims.end(), pcvar_dims_.begin());
}

void MeshVInfo::get_mesh_vinfo(MBMesh *&mesh, moab::Interface *&mb_core, nc_type &data_type, size_t &var_chunk_sz, std::vector<pargal::pcdim> &pcvar_dims, std::string &pcvar_loc, double &missing_val, const std::vector<size_t> &var_dims) const
{
	mesh = mesh_;
	mb_core = mb_core_;
	data_type = data_type_;
	var_chunk_sz = var_chunk_sz_;
	pcvar_loc = pcvar_loc_;
	missing_val = pcvar_missing_val_;

	pcvar_dims.resize(pcvar_dims_.size());
	std::vector<pargal::pcdim>::const_reverse_iterator iriter = pcvar_dims_.rbegin();
	std::vector<pargal::pcdim>::reverse_iterator oriter = pcvar_dims.rbegin();

	for(std::vector<size_t>::const_reverse_iterator var_riter = var_dims.rbegin(); (oriter != pcvar_dims.rend()) && (iriter != pcvar_dims_.rend()) && (var_riter != var_dims.rend()); ++oriter, ++iriter, ++var_riter){
		*oriter = *iriter;
		(*oriter).set_size(*var_riter);
	}
	for(; (oriter != pcvar_dims.rend()) && (iriter != pcvar_dims_.rend()); ++oriter, ++iriter){
		*oriter = *iriter;
		(*oriter).set_size(1);
	}
}

} //namespace parncl
