/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef ScdMBMesh_hpp
#define ScdMBMesh_hpp

#include "MBMesh.hpp"
#include "moab/ScdInterface.hpp"
#include "moab/HomXform.hpp"
#include "moab/EntityType.hpp"

namespace parncl{

class ScdMBMesh : public MBMesh{
public:
	virtual ~ScdMBMesh();

	virtual bool ctest(const std::vector<size_t>& var_dims, std::string &pcal_val_loc);
	virtual bool ctest(ParCALMeshVal &val, NclSelectionRecord *sel);
	virtual void conv_to_local(const NclSelectionRecord *gsel, NclSelectionRecord *lsel, const std::vector<size_t> &gdim_sz, std::vector<size_t> &ldim_sz, const std::string &pcal_val_loc);
	/* Conform global sel to mesh conforming values */
	virtual void conform_mesh(NclSelectionRecord *gsel, const std::string &pcal_val_loc);
	/* Conform a vector of global dim sizes to mesh conforming values */
	virtual void conform_mesh(std::vector<size_t> &gdim_sz, const std::string &pcal_val_loc);
	/* Conform a vector of pcvar pcdims to mesh conforming values */
	virtual void conform_mesh(std::vector<pargal::pcdim> &gdims, const std::string &pcal_val_loc);

	/* Get I/J conform dims */
//	virtual std::string get_itag(const std::string &pcal_val_loc) = 0;
//	virtual std::string get_jtag(const std::string &pcal_val_loc ) = 0;
	virtual const std::vector<std::pair<size_t, size_t> >	get_gij_conform_dims(const std::string &pcal_val_loc) = 0;	
	virtual const std::vector<std::pair<size_t, size_t> >	get_lij_conform_dims(const std::string &pcal_val_loc) = 0;

protected:
	ScdMBMesh(moab::Interface *mb_gcore, MPI_Comm mb_comm, moab::ParallelComm *mb_pcomm, MeshFInfo *mesh_finfo, const std::string &part_method);

private:
	ScdMBMesh();
	void dbg_print_scd_box(moab::ScdBox *) const;
	void find_scd_box(void );
	void init_ijk_info(void );
	moab::ScdInterface *scdi_;
	moab::ScdBox *scd_box_;
//	std::vector<std::pair<size_t, size_t> > gij_conform_dims_;
//	std::vector<std::pair<size_t, size_t> > lijk_dims_;
	std::vector<std::pair<double, double> > lijk_coords_min_max_;
};

} // namespace parncl

#endif // ScdMBMesh_hpp
